--ͻ
--   ۻ  ۻ  ۻ    ۻ ۻ  ۻ ۻ     ۻ ۻ    
--  ͼ ۻ ۻۺ ۻ ۺ ۺ     ͼ ۻ   
--  ۺ      ۺ   ۺ ۺ ɼ ۺ ۺ     ۻ   ɼ   
--  ۺ      ۺ   ۺ ۺͼۺ ͼ  ۺ ۺ     ͼ   ۻ   
--  ۻ ɼ ۺ    ۺ ۺ      ۺ ۻ ۻ ۺ  ۺ   
--   ͼ  ͼ  ͼ    ͼ ͼ      ͼ ͼ ͼ ͼ  ͼ   
--͹
--        (C) 1993-1998, Detlef Reimers, Hamburg, dreimers@aol.com         
--͹
-- Dieses Programm ist ein einfacher Compiler fr mathematische Terme. Es  
-- knnen einfache Ziffern als Zahlen eingegeben werden und +, -, *, / als 
-- Operatoren. Klammern werden ebenfalls untersttzt. Der Compiler erzeugt 
-- 68000 Assembler-Code. Ein- und Ausgabe erfolgen auf dem Bildschirm. Bei 
-- Fehlern im Term wird das Programm mit einer Fehlermeldung beendet.      
--                                                                         
-- Es folgt die Beschreibung der Grammatik der bersetzten mathematischen  
-- Terme in Backus-Naur Form oder kurz BNF:                                
--                                                                         
--      EXPRESSION := TERM | TERM + EXPRESSION | TERM - EXPRESSION         
--            TERM := FACTOR | FACTOR * TERM | FACTOR / TERM               
--          FACTOR := ( EXPRESSION ) | DIGIT                               
--           DIGIT := 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9                
--ͼ
global constant TAB = 9, LF = 10

atom look
integer expr, lcount

--Ŀ
-- Ein neues Zeichen von der Eingabe lesen.                                
--
procedure Getchar()
  
  look = getc(0)
  
end procedure

--Ŀ
-- Eine Fehlermeldung ausgeben.                                            
--
procedure Error(sequence s)
  
  printf(1,"\n\tError: %s", {s})
  
end procedure

--Ŀ
-- Fehlermeldung ausgeben und abbrechen.                                   
--
procedure Abort(sequence s)
  
  Error(s)
  abort(0)
  
end procedure

--Ŀ
-- Ausgeben, was erwartet wurde.                                           
--
procedure Expected(sequence s)
  
  Abort(s & " expected")
  
end procedure

--Ŀ
-- In Kleinbuchstaben umwandeln.                                           
--
global function Lower(atom x)

  return x + (x >= 'A' and x <= 'Z') * ('a' - 'A')
end function

--Ŀ
-- In Grobuchstaben umwandeln.                                            
--
global function Upper(atom x)

  return x - (x >= 'a' and x <= 'z') * ('a' - 'A')
end function

--Ŀ
-- Ein Alpha-Zeichen erkennen.                                             
--
function Isalpha(atom c)
  
  atom a
  a = Upper(c)
  
  return a >= 'A' and a <= 'Z'
end function

--Ŀ
-- Eine Dezimal-Ziffer erkennen.                                           
--
function Isdigit(atom c)
  
  return c >= '0' and c <= '9'
end function

--Ŀ
-- Ein alphanumerisches Zeichen erkennen.                                  
--
function Isalnum(atom c)
  
  return Isalpha(c) or Isdigit(c)
end function

--Ŀ
-- Einen Additionsoperator erkennen.                                       
--
function Isaddop(atom c)
  
  return c = '+' or c = '-'
end function

--Ŀ
-- Einen Multiplikationsoperator erkennen.                                 
--
function Ismulop(atom c)
  
  return c = '*' or c = '/'
end function

--Ŀ
-- Leerzeichen erkennen.                                                   
--
function Iswhite(atom c)
  
  return c = ' ' or c = TAB
end function

--Ŀ
-- Fhrende Leerzeichen berspringen.                                      
--
procedure Skipwhite()
  
  while Iswhite(look) do
    Getchar()
  end while

end procedure

--Ŀ
-- Ein spezielles Eingabezeichen prfen.                                   
--
procedure Match(atom x)
  
  if look != x then
    Expected("\"" & x & "\"")
  else
    Getchar()
    Skipwhite()
  end if
  
end procedure

--Ŀ
-- Einen Bezeichner einlesen.                                              
--
function Getname()
  
  sequence token, temp
  
  token = ""
  
  if not Isalpha(look) then
    Expected("Name")
  end if
  
  while Isalnum(look) do
    token = token & Upper(look)
    Getchar()
  end while
  
  temp = token
  Skipwhite()

  return temp 
end function

--Ŀ
-- Eine Zahl einlesen.                                                     
--
function Getnum()
  
  sequence value, temp
  
  value = ""
  
  if not Isdigit(look) then
    Expected("Integer")
  end if
  
  while Isdigit(look) do
    value = value & look
    Getchar()
  end while
  
  temp = value
  Skipwhite()

  return temp
end function

--Ŀ
-- Einen String mit Tab ausgeben.                                          
--
procedure Emit(sequence s)
  
  printf(1, "\t%s", {s})
  
end procedure

--Ŀ
-- Einen String mit Tab und CRLF ausgeben.                                 
--
procedure Emitln(sequence s)
  
  printf(1, "\t%s\n", {s})
  
end procedure

--Ŀ
-- Einen Bezeichner parsen und bersetzen.                                 
--
procedure Ident()
  
  sequence name
  
  name = Getname()
  
  if look = '(' then
    Match('(')
    Match(')')
    Emitln("BSR   " & name)
  else
    Emitln("MOVE  " & name & "(PC),D0")
  end if
  
end procedure

--Ŀ
-- Einen mathematischen Faktor parsen und bersetzen. Faktor ruft rekursiv 
-- Expression auf. Da beide Prozeduren sich gegenseitig aufrufen knnen,   
-- mu eine Vorwrtsreferenz mit expr auf Expression implementiert werden. 
--
procedure Factor()
  
  if look = '(' then
    Match('(')
    call_proc(expr,{})
    Match(')')
  elsif Isalpha(look) then
    Ident()
  else
    Emitln("MOVE  #" & Getnum() & ",D0")
  end if
  
end procedure

--Ŀ
-- Eine Multiplikation erkennen und bersetzen.                            
--
procedure Multiply()
  
  Match('*')
  Factor()
  Emitln("MULS  (SP)+,D0")
  
end procedure

--Ŀ
-- Eine Division erkennen und bersetzen.                                  
--
procedure Divide()
  
  Match('/')
  Factor()
  Emitln("MOVE  (SP)+,D1")
  Emitln("EXT.L D0")
  Emitln("DIVS  D1,D0")
  
end procedure

--Ŀ
-- Einen mathematischen Term parsen und bersetzen.                        
--
procedure Term()
  
  Factor()
  
  while Ismulop(look) do
    Emitln("MOVE  D0,-(SP)")
    if look = '*' then
      Multiply()
    elsif look = '/' then
      Divide()
    end if
  end while
  
end procedure

--Ŀ
-- Eine Addition erkennen und bersetzen.                                  
--
procedure Add()
  
  Match('+')
  Term()
  Emitln("ADD   (SP)+,D0")
  
end procedure

--Ŀ
-- Eine Subtraktion erkennen und bersetzen.                               
--
procedure Subtract()
  
  Match('-')
  Term()
  Emitln("SUB   (SP)+,D0")
  Emitln("NEG   D0")
  
end procedure

--Ŀ
-- Einen Ausdruck parsen und bersetzen.                                   
--
procedure Expression()
  
  if Isaddop(look) then
    Emitln("CLR   D0")
  else
    Term()
  end if
  
  while Isaddop(look) do
    Emitln("MOVE  D0,-(SP)")
    if look = '+' then
      Add()
    elsif look = '-' then
      Subtract()
    end if
  end while
  
end procedure

--Ŀ
-- Einen Zuordnungsbefehl parsen und bersetzen.                           
--
procedure Assignment()
  
  sequence name
  
  puts(1, "\n")
  name = Getname()
  Match('=')
  Expression()
  Emitln("LEA   " & name & "(PC),A0")
  Emitln("MOVE  D0,(A0)")
  
end procedure

--Ŀ
-- Initialisierung.                                                        
--
procedure Init()
  
  Getchar()
  Skipwhite()
  
end procedure

--Ŀ
-- Das Hauptprogramm: Zu Beginn wird die Prozedur-Nr von Expression        
-- bestimmt. Diese wird fr die Vorwrtsreferenz in Factor bentigt.       
--
expr = routine_id("Expression") 
Init()
Assignment()

if look != LF then
  Expected("Newline")
end if  

abort(0)

--
