--ͻ
--   ۻ ۻ  ۻ ۻ ۻ ۻ  ۻ  ۻ  ۻ   
--   ۺ ۻۺ ͼ ͼ ۻ ۻ ۻ ͼ   
--   ۺ ۺ    ۺ    ۻ   ɼ ɼ ɼ ۻ     
--   ۺ ۺۺ    ۺ    ͼ   ۻ ͼ  ۻ ͼ     
--   ۺ ۺ  ۺ    ۺ    ۻ ۺ  ۺ ۺ      ۺ  ۺ ۻ   
--   ͼ ͼ  ͼ    ͼ    ͼ ͼ  ͼ ͼ      ͼ  ͼ ͼ   
--͹
--        (C) 1993-1998, Detlef Reimers, Hamburg, dreimers@aol.com         
--͹
-- Dieses Programm ist ein einfacher Interpreter fr mathematische Terme.  
-- Es knnen ganze Zahlen und +, -, *, / als Operatoren eingegeben werden. 
-- Das unre Plus und Minus wie auch Klammerterme knnen benutzt werden.   
-- Die Ein- und Ausgabe erfolgt ber den Bildschrm. Bei Auftreten von      
-- Fehlern im Term wird das Programm mit einer Fehlermeldung beendet.      
-- Es folgt die Beschreibung der verwendeten Grammatik fr mathematische   
-- Terme in Backus-Naur Form oder kurz BNF:                                
--                                                                         
--      EXPRESSION := TERM | TERM + EXPRESSION | TERM - EXPRESSION         
--            TERM := FACTOR | FACTOR * TERM | FACTOR / TERM               
--          FACTOR := ( EXPRESSION ) | INTEGER                             
--         INTEGER := DIGIT | [ DIGIT ]                                    
--           DIGIT := 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9                
--ͼ
--ͻ
--                     BEDIENUNG DES INTERPRETERS:                         
--                                                                         
-- Die Eingabe erfolgt mit ?<Term><Wert> oder mit <Term>=<Wert>. Als       
-- Variablen knnen alle Buchstaben des Alphabets benutzt werden. Z.B.:    
--                                                                         
--      ?a120                                  a <- 120                    
--      a=a-90                                 a <-  30                    
--      b=25                                   b <-  25                    
--      ?xb+5*(40-a)                           x <-  75                    
--      x=5*a-b                                x <- 125                    
--                                                                         
-- Die Ausgabe erfolgt mit !<Term>. Bei Auftreten des Zeichens <.> wird    
-- das Programm beendet. Nach Fehlern wird eine Fehlermeldung angezeigt    
-- und das Programm bricht ab. Z.B.:                                       
--                                                                         
--      a=240                                  a <- 240                    
--      b=-40                                  b <- -40                    
--      c=a+5*b                                c <-  40                    
--      !c                                    40 -> CRT                    
--      !a+2*(a+4*b)                         400 -> CRT                    
--      !-b.                                  40 -> CRT und Ende           
--                                                                         
--ͼ
include get.e
include graphics.e

--Ŀ
-- Globale Konstanten.                                                     
--
global constant
TAB = 9, LF = 10

--Ŀ
-- Variablen.                                                              
--
atom look, expr
sequence table, pos

--Ŀ
-- Den Variablen-Bereich initialisieren (a bis z).                         
--
procedure Inittable()
  
  table = repeat(0,'Z' - 'A' + 1)
  
end procedure

--Ŀ
-- Ein neues Zeichen von der Eingabe einlesen.                             
--
procedure Getchar()
  
  look = getc(0)
  pos[2] = pos[2] + 1
  
end procedure

--Ŀ
-- Eine Fehlermeldung ausgeben.                                            
--
procedure Error(sequence s)
  
  puts(1,"\nError: " & s & "\n")
  
end procedure

--Ŀ
-- Fehlermeldung ausgeben und abbrechen.                                   
--
procedure Abort(sequence s)
  
  Error(s)
  abort(0)
  
end procedure

--Ŀ
-- Ausgeben, was erwartet wurde.                                           
--
procedure Expected(sequence s)
  
  Abort(s & " expected")
  
end procedure

--Ŀ
-- Ein spezielles Eingabezeichen prfen.                                   
--
procedure Match(atom c)
  
  if look = c then
    Getchar()
  else
    Expected("\"" & c & "\"")
  end if
  
end procedure

--Ŀ
-- Erkenne und berspringe ein Neue-Zeile-Zeichen.                         
--
procedure Newline()
  
  if look = LF then
    puts(1, "\n")
    Getchar()
  end if
  
  pos = get_position()
  
end procedure

--Ŀ
-- In Kleinbuchstaben umwandeln.                                           
--
global function Lower(atom x)

    return x + (x >= 'A' and x <= 'Z') * ('a' - 'A')
end function

--Ŀ
-- In Grobuchstaben umwandeln.                                            
--
global function Upper(atom x)

    return x - (x >= 'a' and x <= 'z') * ('a' - 'A')
end function

--Ŀ
-- Ein Alpha-Zeichen erkennen.                                             
--
function Isalpha(atom c)
  
  atom a
  a = Upper(c)
  
  return a >= 'A' and a <= 'Z'
end function

--Ŀ
-- Eine Dezimal-Ziffer erkennen.                                           
--
function Isdigit(atom c)
  
  return c >= '0' and c <= '9'
end function

--Ŀ
-- Einen Additionsoperator erkennen.                                       
--
function Isaddop(atom c)
  
  return c = '+' or c = '-'
end function

--Ŀ
-- Einen Multiplikationsoperator erkennen.                                 
--
function Ismulop(atom c)
  
  return c = '*' or c = '/'
end function

--Ŀ
-- Einen Bezeichner einlesen.                                              
--
function Getname()
  
  atom a
  
  if not Isalpha(look) then
    Expected("Name")
  end if
  
  a = Upper(look)
  Getchar()

  return a 
end function

--Ŀ
-- Eine Zahl einlesen.                                                     
--
function Getnum()
  
  atom value
  
  value = 0
  
  if not Isdigit(look) then
    Expected("Integer")
  end if
  
  while Isdigit(look) do
    value = 10 * value + look - '0'
    Getchar()
  end while

  return value
end function

--Ŀ
-- Einen mathematischen Faktor parsen und bersetzen. Faktor ruft rekursiv 
-- Expression auf. Da beide Prozeduren sich gegenseitig aufrufen knnen,   
-- mu eine Vorwrtsreferenz mit expr auf Expression implementiert werden. 
--
function Factor()

  atom value
  
  if look = '(' then
    Match('(')
    value = call_func(expr,{})
    Match(')')
  elsif Isalpha(look) then
    value = table[Getname()-'@']
  else
    value = Getnum()
  end if
  
  return value
end function

--Ŀ
-- Einen mathematischen Term parsen und bersetzen.                        
--
function Term()

  atom value, temp
  
  value = Factor()
  
  while Ismulop(look) do
    if look = '*' then
      Match('*')
      value = value * Factor()
    elsif look = '/' then
      Match('/')
      temp = Factor()
      if temp = 0 then
        Abort("Division by Zero")
      end if
      value = floor(value / temp)
    end if
  end while
  
  return value
end function

--Ŀ
-- Einen Ausdruck parsen und bersetzen.                                   
--
function Expression()

  atom value
  
  if Isaddop(look) then
    value = 0
  else
    value = Term()
  end if
  
  while Isaddop(look) do
    if look = '+' then
      Match('+')
      value = value + Term()
    elsif look = '-' then
      Match('-')
      value = value - Term()
    end if
  end while
  
  return value
end function

--Ŀ
-- Eingabe-Routine.                                                        
--
procedure Input()
  
  Match('?')
  table[Getname()-'@'] = Expression()

end procedure

--Ŀ
-- Ausgabe-Routine.                                                        
--
procedure Output()
  
  atom n
  Match('!')
  n = Expression()
  position(pos[1], pos[2])
  printf(1, "=%g", n)

end procedure

--Ŀ
-- Einen Zuordnungsbefehl parsen und bersetzen.                           
--
procedure Assignment()
  
  atom name
  
  name = Getname()
  Match('=')
  table[name-'@'] = Expression()
  
end procedure

--Ŀ
-- Initialisierung durch Lesen des ersten Eingabezeichens.                 
--
procedure Init()
  
  puts(1,"Term-Interpreter\n")
  pos = {1, 0}
  Inittable()
  Getchar()
  
end procedure

--Ŀ
-- Das Hauptprogramm: Zu Beginn wird die Prozedur-Nr von Expression        
-- bestimmt. Diese wird fr die Vorwrtsreferenz in Factor bentigt.       
-- Das Programm endet nach einem Fehler oder nach Eingabe von <.>.         
--
expr = routine_id("Expression") 

Init()

while look != '.' do
  if look = '?' then
    Input()
  elsif look = '!' then
    Output()
  else
    Assignment()
  end if
  
  Newline()
end while

--
